package com.xylink.util;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.base.Predicate;
import com.xylink.config.SDKConfig;
import com.xylink.config.SDKConfigMgr;
import com.xylink.log.XyLinkLog;
import com.xylink.model.HttpMethod;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.StringUtils;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import java.io.*;
import java.net.HttpURLConnection;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.URL;
import java.security.cert.X509Certificate;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

/**
 * Created by maolizhi on 12/13/2016.
 */
public class HttpUtil {

    private static XyLinkLog logger = SDKConfigMgr.getXyLinkLog();

    private static String proxyHost;
    private static int proxyPort;
    private static Integer connectTimeout;
    private static Integer readTimeout;

    public static Result getResponse(String surl, String method, String jsonEntity, Class<?> clazz) throws IOException {
        return getResponse(surl, method, jsonEntity, clazz, null);
    }
    public static Result getResponse(String surl, String method, String jsonEntity, Class<?> clazz, String clientId) throws IOException {
        final SDKConfig sdkConfig = SDKConfigMgr.getSdkConfig(surl,clientId);
        //请求发起时间
        final long currentTimeMillis = System.currentTimeMillis();
        //before 处理
        beforeCheck(sdkConfig, currentTimeMillis);

        //主业务处理
        Result result = executeHttp(surl, method, jsonEntity, clazz, sdkConfig);

        //after 处理，判定是否进行异常重试，和上面的before对应
        return afterCheck(surl, method, jsonEntity, clazz, sdkConfig, currentTimeMillis, result);
    }

    private static Result afterCheck(String surl, String method, String jsonEntity, Class<?> clazz, SDKConfig sdkConfig, final long currentTimeMillis, Result result) throws IOException {
        if (!result.isSuccess()
                && result.getData() != null
                && result.getData() instanceof RestMessage
                && sdkConfig.getSignatureVersion().isNewSign()
                //token错误
                && (10002000 == ((RestMessage) result.getData()).getErrorCode()
                //时间不对
                || (10002015 == ((RestMessage) result.getData()).getErrorCode() && sdkConfig.getSignatureVersion().isTimeSafeVersion()))
        ) {
            //如果在请求发送后sdkconfig有更新，就可以直接进行再次请求，不用再获取了
            SDKConfigMgr.initAccesstoken(sdkConfig, new Predicate<SDKConfig>() {
                @Override
                public boolean apply(SDKConfig config) {return currentTimeMillis < config.getUpdateTime();}
            }, false);

            return executeHttp(surl, method, jsonEntity, clazz, sdkConfig);
        }
        return result;
    }

    private static void beforeCheck(final SDKConfig sdkConfig, final long currentTimeMillis) {
        if (sdkConfig.getSignatureVersion().isNewSign()) {
            //新签名的情况下需要进行before处理，检查accesstoken的过期时间，判断是否进行处理
            //判断内容分为两部分，1、是否快过期（过期时间低于5m）则异步刷新accessToken 2、是否已过期，则同步刷新accessToken,2s的冗余时间
            if (sdkConfig.getAccessTokenExpireIn() - currentTimeMillis < 2000) {
                //过期,给提前2s进行请求,同步刷新
                SDKConfigMgr.initAccesstoken(sdkConfig, new Predicate<SDKConfig>() {
                    @Override
                    public boolean apply(SDKConfig config) {
                        //检查过期时间是否大于5分钟，新刷新出来的token过期时间必须大于5分钟，大于5分钟的认为是新刷新出来的token，可以直接使用
                        return config.getAccessTokenExpireIn() - currentTimeMillis > 300000;
                    }
                }, false);

            } else if (sdkConfig.getAccessTokenExpireIn() - currentTimeMillis < 300000) {
                //将要5分钟后过期，异步立即刷新
                //抢锁处理，任何一个抢到了就可以进行刷新，其他的直接退出执行
                SchedulerUtil.schedule(new Runnable() {
                    @Override
                    public void run() {
                        SDKConfigMgr.initAccesstoken(sdkConfig, new Predicate<SDKConfig>() {
                            public boolean apply(SDKConfig config) {
                                //检查过期时间是否大于5分钟，新刷新出来的token过期时间必须大于5分钟，大于5分钟的认为是新刷新出来的token，可以直接使用
                                return config.getAccessTokenExpireIn() - currentTimeMillis > 300000;
                            }
                        }, true);
                    }
                }, 0, TimeUnit.MILLISECONDS);
            }
        }
    }

    public static Result executeHttp(String surl, String method, String jsonEntity, Class<?> clazz, SDKConfig sdkConfig) throws IOException {
       return executeHttp(surl, method, jsonEntity, clazz, sdkConfig, true);
    }

    public static Result executeHttp(String surl, String method, String jsonEntity, Class<?> clazz, SDKConfig sdkConfig, boolean printErrorLog) throws IOException {
        HttpURLConnection conn = HttpUtil.getHttpURLConnection(jsonEntity, surl, method, sdkConfig);
        InputStream is;
        int status = conn.getResponseCode();
        if (status == 200 || status == 204) {
            is = conn.getInputStream();
        } else {
            is = conn.getErrorStream();
        }

        String respData = HttpUtil.getResponseData(is);
        //错误日志进行详细的日志输出
        if (status == 200 || status == 204) {
            logger.info("xylink request " +
                    "url:" + conn.getURL() + ", " +
                    "method:" + method + ", " +
                    "jsonEntity:" + jsonEntity +
                    "\n response:" + respData);
        } else if (printErrorLog){
            logger.error("xylink request " +
                    "url:" + conn.getURL() + ", " +
                    "method:" + method + ", " +
                    "header:" + conn.getHeaderFields() + ", " +
                    "jsonEntity:" + jsonEntity +
                    "sdkConfig:" + sdkConfig +
                    "\n response:" + respData);
        }else {
            logger.info("xylink request " +
                    "url:" + conn.getURL() + ", " +
                    "method:" + method + ", " +
                    "header:" + conn.getHeaderFields() + ", " +
                    "jsonEntity:" + jsonEntity +
                    "sdkConfig:" + sdkConfig +
                    "\n response:" + respData);
        }
        conn.disconnect();
        Result result = getResult(status, respData, clazz, printErrorLog);
        return result;
    }

    public static Result getByteStreamResponse(String surl, String method, String jsonEntity) throws IOException {
        return getByteStreamResponse(surl, method, jsonEntity, null);
    }
    public static Result getByteStreamResponse(String surl, String method, String jsonEntity, String clientId) throws IOException {
        SDKConfig sdkConfig = SDKConfigMgr.getSdkConfig(surl, clientId);
        HttpURLConnection conn = HttpUtil.getHttpURLConnection(jsonEntity, surl, method, sdkConfig);
        InputStream is = null;
        int status = conn.getResponseCode();
        if (status == 200 || status == 204) {
            is = conn.getInputStream();
        } else {
            is = conn.getErrorStream();
        }
        byte[] bytes = HttpUtil.getStreamResponse(is);
        conn.disconnect();
        return getByteResult(status, bytes);
    }

    /**
     * convert byte array to image or video
     *
     * @param bytes
     * @param path
     */
    public static void convertByteArrayToImageOrVideo(byte[] bytes, String path) {
        FileOutputStream fos = null;
        try {
            fos = new FileOutputStream(path);
            fos.write(bytes);
        } catch (FileNotFoundException e) {
            logger.error("File Path not found!", e);
            //System.out.println("File Path not found!");
        } catch (IOException e) {
            logger.error("Convert byte array to image,IO Errot!", e);
            //System.out.println("Convert byte array to image,IO Errot!");
        }
    }


    private static HttpURLConnection getHttpURLConnection(String jsonEntity, String surl, String method, SDKConfig sdkConfig) throws IOException {
        URL url = new URL(surl);
        HttpURLConnection conn = null;
        if (StringUtils.isNotBlank(proxyHost)) {
            logger.info("use proxy");
            //System.out.println("use proxy");
            Proxy proxy = new Proxy(Proxy.Type.HTTP, new InetSocketAddress(proxyHost, proxyPort));
            conn = (HttpURLConnection) url.openConnection(proxy);
        } else {
            conn = (HttpURLConnection) url.openConnection();
        }
        if (sdkConfig != null && sdkConfig.getSignatureVersion().isNewSign()) {
            String uri = url.getPath();
            if (!uri.startsWith("/admin/login") && !uri.startsWith("/api/rest/uaa/external/sysinfo")) {
                HashMap globalParams = new HashMap<String, String>();
                globalParams.put(SignUtil.SIGN_CLIENT_ID_KEY_NEW, sdkConfig.getXyClientid());
                globalParams.put(SignUtil.SIGN_SIGN_TYPE_KEY_NEW, sdkConfig.getSignType().name());
                String timeStamp = String.valueOf(System.currentTimeMillis() + sdkConfig.getTimeLag());
                globalParams.put(SignUtil.SIGN_TIMESTAMP_KEY_NEW, timeStamp);
                String nonce = DigestUtils.sha256Hex(timeStamp + SignUtil.randomAlphanumeric(16));
                globalParams.put(SignUtil.SIGN_NONCE_KEY_NEW, nonce);
                //普通接口和获取token接口的签名方式
                String signToken = uri.startsWith("/api/rest/gw/external/app_token/v1") ? sdkConfig.getXyClientSecret() : sdkConfig.getXySignSecret();
                String query = StringUtils.isEmpty(url.getQuery()) ? "" : "?" + url.getQuery();
                String sign = SignUtil.getSign(method, globalParams, uri + query, jsonEntity, signToken);
                conn.addRequestProperty(SignUtil.SIGN_CLIENT_ID_KEY_NEW, sdkConfig.getXyClientid());
                conn.addRequestProperty(SignUtil.SIGN_SIGN_TYPE_KEY_NEW, sdkConfig.getSignType().name());
                conn.addRequestProperty(SignUtil.SIGN_TIMESTAMP_KEY_NEW, timeStamp);
                conn.addRequestProperty(SignUtil.SIGN_NONCE_KEY_NEW, nonce);
                conn.addRequestProperty(SignUtil.SIGN_SIGN_KEY_NEW, sign);
                conn.addRequestProperty(SignUtil.X_XY_SIGNATUREVERSION, sdkConfig.getSignatureVersion().getVersion());
                if (!uri.startsWith("/api/rest/gw/external/app_token/v1")){
                    conn.addRequestProperty(SignUtil.Authorization, "Bearer " + sdkConfig.getAccessToken());
                }
            }
        }

        conn.setRequestMethod(method);
        if (sdkConfig != null && sdkConfig.isEnableSecurityHttp()) {
            if (HttpMethod.PUT.name().equals(method) || HttpMethod.DELETE.name().equals(method)) {
                conn.setRequestMethod(HttpMethod.POST.name());      //PUT和DELETE转换为POST
                conn.addRequestProperty("Xy-Method", method);
                logger.info("enable security https method, source:" + method);
            }
        }
        conn.addRequestProperty("Accept", "application/json");
        conn.addRequestProperty("Accept-Charset", "UTF-8");
        conn.addRequestProperty("Content-Type", "application/json;charset=utf-8");
        Map<String, String> extrasHeader = sdkConfig != null ? sdkConfig.getExtrasHeader() : null;
        if (null != extrasHeader && !extrasHeader.isEmpty()) {
            for (Map.Entry<String, String> entry : extrasHeader.entrySet()) {
                conn.addRequestProperty(entry.getKey(), entry.getValue());
            }
        }
        if (Objects.nonNull(connectTimeout)){
            conn.setConnectTimeout(connectTimeout);
        }
        if (Objects.nonNull(readTimeout)){
            conn.setReadTimeout(readTimeout);
        }
        conn.setDoOutput(true);
        conn.setDoInput(true);
        if (jsonEntity != null && !"".equals(jsonEntity.trim())) {
            conn.getOutputStream().write(jsonEntity.getBytes("utf-8"));
        }
        conn.connect();
        return conn;
    }

    private static String getResponseData(InputStream is) throws IOException {
        if (is != null) {
            Reader reader = new InputStreamReader(is, "UTF-8");
            StringBuilder response = new StringBuilder();
            char[] buff = new char[2048];
            for (int n; (n = reader.read(buff)) != -1; ) {
                response.append(buff, 0, n);
            }
            return response.toString();
        } else {
            return "";
        }
    }


    private static Result getResult(int status, String respData, Class<?> clazz, boolean printErrorLog) {
        ObjectMapper objectMapper = new ObjectMapper();
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        int responseCode = 200;
        RestMessage restMessage = null;
        Result result = new Result();
        if (status == 200 || status == 204) {
            try {
                result.setSuccess(true);
                if (clazz != null) {
                    if (!clazz.getName().equals(String.class.getName())) {
                        if (respData != null && !"".equals(respData.trim()))
                            result.setData(objectMapper.readValue(respData, clazz));
                    } else {
                        result.setData(respData);
                    }
                }
                result.setErrorStatus(responseCode);
            } catch (Exception e) {
                logger.error("JSon String convert to Object error!", e);
                result.setSuccess(false);
                result.setData(null);
                result.setErrorStatus(204);
                result.setErrorMsg("data trans error");
                //System.out.println("JSon String convert to Object error!");
            }
        } else {
            result.setErrorStatus(status);
            result.setSuccess(false);
            try {
                restMessage = objectMapper.readValue(respData, RestMessage.class);
                responseCode = restMessage.getErrorCode();
                result.setData(restMessage);
                result.setErrorStatus(responseCode);// 有详细错误码，使用详细的错误码
                result.setErrorMsg(restMessage.getUserMessage());
            } catch (Exception e) {
                if (printErrorLog){
                    logger.error("JSon String convert to RestMessage error!", e);
                }else {
                    logger.info("JSon String convert to RestMessage error! msg:{}" +  result);
                }
                result.setSuccess(false);
                result.setData(null);
                result.setErrorStatus(400);
                result.setErrorMsg("RestMessage trans error");
                //System.out.println("JSon String convert to RestMessage error!");
            }
        }
        return result;
    }


    private static byte[] getStreamResponse(InputStream is) throws IOException {
        if (is != null) {
            ByteArrayOutputStream swapStream = new ByteArrayOutputStream();
            byte[] buff = new byte[1024]; //buff用于存放循环读取的临时数据
            int rc = 0;
            while ((rc = is.read(buff, 0, 100)) > 0) {
                swapStream.write(buff, 0, rc);
            }
            byte[] bytes = swapStream.toByteArray(); //bytes为转换之后的结果
            return bytes;
        }
        return null;

    }

    private static Result getByteResult(int status, byte[] bytes) {
        ObjectMapper objectMapper = new ObjectMapper();
        int responseCode = 200;
        RestMessage restMessage = null;
        Result result = new Result();
        if (status == 200 || status == 204) {
            result.setSuccess(true);
            result.setData(bytes);
            result.setErrorStatus(200);

        } else {
            try {
                String respData = new String(bytes);
                restMessage = objectMapper.readValue(respData, RestMessage.class);
                responseCode = restMessage.getErrorCode();
                result.setSuccess(false);
                result.setData(restMessage);
                result.setErrorStatus(responseCode);
            } catch (Exception e) {
                logger.error("JSon String convert to RestMessage error!", e);
                //System.out.println("JSon String convert to RestMessage error!");
            }
        }

        return result;
    }

    public static void setProxy(String host, int port) {
        proxyHost = host;
        proxyPort = port;
    }

    public static void setConnectTimeout(Integer connectTimeout) {
        HttpUtil.connectTimeout = connectTimeout;
    }

    public static void setReadTimeout(Integer readTimeout) {
        HttpUtil.readTimeout = readTimeout;
    }

//    static {
//        try {
//            // 创建信任所有证书的TrustManager
//            TrustManager[] trustAllCerts = new TrustManager[]{
//                    new X509TrustManager() {
//                        public X509Certificate[] getAcceptedIssuers() {
//                            return null;
//                        }
//
//                        public void checkClientTrusted(X509Certificate[] certs, String authType) {
//                        }
//
//                        public void checkServerTrusted(X509Certificate[] certs, String authType) {
//                        }
//                    }
//            };
//
//            // 初始化SSLContext并设置TrustManager
//            SSLContext sc = SSLContext.getInstance("TLS");
//            sc.init(null, trustAllCerts, null);
//
//            // 获取SSLSocketFactory并设置到HttpsURLConnection中
//            HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());
//
//            // 设置主机名验证器（可选）
//            HttpsURLConnection.setDefaultHostnameVerifier((hostname, session) -> true);
//
//        } catch (Exception e) {
//            logger.error("ignore TrustManager failed ", e);
//        }
//    }

}
